/**
 * This is the Message class.
 */

package salsa.language;

public class Message implements java.io.Serializable {
	/**
		Actor that the message is being sent from.
	*/
	private Actor sender;

	/**
		Actor that the message is being sent to.
	*/
	private Actor target;

	/**
		Name of the method that should be invoked.
	*/
	private String methodName;

	/**
		Array of values that are copied to be the arguments of the method. If one of
		the arguments is the value of a yet to be evaluated token an instance of the
		Token class should be subituted for it until it is evaluated.
	*/
	private Object[] arguments;

	/**
		Reference to the next message in the continuation.
	*/
	private Message continuation;

	/**
	 * This is the number of times the message has been passed through a forwarder.
	 * If the number reaches a certain threshold, the nameserver should be queried 
	 * to attempt to resolve the actor's new location.
	 */
	private int forwarded;

	/**
	 * Public timestamps for measuring message throughput.
	 * 
	 * Use -Ddebug for output.
	 */
	public long messageCreated;
	public long messageSent;
	public long messageQueued;
	public long messageStarted;
	public long messageFinished;
	
	/**
		This constructor takes a sender, and three arrays containing respective targets,
		message names and argument arrays.
		The constructor creates a linked list of messages for each message in the continuation.
	*/
	public Message( Actor sender, Actor[] targets, String[] methodNames, Object[][] arguments) {

		this( sender, targets[ 0 ], methodNames[ 0 ], arguments[ 0 ] );

		Message m = this;

		// Record the time.
		messageCreated = System.currentTimeMillis();

		for (int i=1;i<targets.length;++i) {
			m.continuation = new Message( targets[ i-1 ], targets[ i ], methodNames[ i ], arguments[ i ] );
			m = m.continuation;
		}
		
		forwarded = 0;
	}

	/**
		The constructor takes a sender, target, message name, and array of arguments and
		constructs one Message object. The continuation field remains null.
	*/
	public Message( Actor sender, Actor target, String methodName, Object[] arguments) {
		this.sender = sender;
		this.target = target;
		this.methodName = methodName;
		this.arguments = arguments;
		
		forwarded = 0;
	}

	/**
	 * @return int
	 */
	public int getForwarded() {
		return forwarded;
	}

	/**
	 * Sets the forwarded.
	 * @param forwarded The forwarded to set
	 */
	public void setForwarded(int forwarded) {
		this.forwarded = forwarded;
	}

	/**
		This is a simple way to output the name of a message in a
		readable format. used for testing purposes.
	*/
	public String getName() {
		String text = "[Source: ";
		if (sender != null) text += sender.getName();
		else text += "null";

		text += ", Target: ";
		if (target != null) text += target.getName();
		else text += "null";

		text += ", Method: " + methodName;

		if (arguments != null) {
			text += ", Arguments: {";
			for (int i = 0; i < arguments.length; i++) {
				if (arguments[i] != null) text += arguments[i].getClass().getName();
				else text += "null";

				if (i != (arguments.length-1)) text += ", ";
			}
			text += "}";
		}
		else text += null;

//		if (continuation != null) text +=  ", Continuation: " + continuation.getName();

		if (continuation != null) text += ", Continuation: true";
		else text += ", Continuation: false";

		text += "]";
		return text;
	}

	/**
		This method sets any arguments which are instances of Token to the specified reference.
	*/
	public void updateTokens( Object newToken ) {
		for (int i=0; i<arguments.length; ++i) {
			if (arguments[i] instanceof Token) arguments[i] = newToken;
		}
	}

	/**
		This method returns the target of the message
	*/
	public Actor getTarget() {
		return target;	
	}

	/**
		This method returns the source of the message
	*/
	public Actor getSource() {
		return sender;	
	}

	/**
		This method returns the arguments of the message.
	*/
	public Object[] getArguments() {
		return arguments;
	}

	/**
		This method returns the continuation of the message or null
		if none exists.
	*/
	public Message getContinuation() {
		return continuation;
	}

	/**
		This method is used when a migration occurs,
		after the continuation is prioritySent to the
		Actor.
	*/
	public void removeContinuation() {
		continuation = null;
	}

	/**
		This recursive method is called by the MessageGroup class that
		implements the Join Director. It adds the specified message to the
		end of the continuation list.
	*/
	protected void addContinuation( Message newContinuation ) {

		if (continuation == null) {
			continuation = newContinuation;

		} else {
			continuation.addContinuation( newContinuation );
		}
	}

	/**
		This method returns the name of the message.
	*/
	public String getMethodName() {
		return methodName;
	}

	/**
		This method inserts a current continuation into a message, to allow
		for first class continuations
	*/
	public void updateFirstClass(Message firstClassContinuation) {
		if (firstClassContinuation == null) {
			return;
		} else if (arguments[0] instanceof FirstClassToken) {
			Message currentContinuation = continuation;

			sender = firstClassContinuation.sender; 
			target = firstClassContinuation.target; 
			methodName = firstClassContinuation.methodName; 
			arguments = firstClassContinuation.arguments;

			Message m = this;
			Message tempContinuation = firstClassContinuation.continuation;

			while (tempContinuation != null) {
				m.continuation = new Message( m.target, 
							      tempContinuation.target, 
							      tempContinuation.methodName, 
							      tempContinuation.arguments );
				m = m.continuation;
				tempContinuation = tempContinuation.continuation;
			}

			if (currentContinuation != null) {
				m = currentContinuation;
				m.updateFirstClass(firstClassContinuation);
			}
		} else if (continuation != null) {
			continuation.updateFirstClass(firstClassContinuation);
		}
	}	
}




